/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.bridge;

import java.io.IOException;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.HashMap;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xquark.mapper.dbms.AbstractConnection;
import org.xquark.mapper.mapping.Loader;
import org.xquark.mapper.mapping.MappingFactory;
import org.xquark.mapper.mapping.RepositoryMapping;
import org.xquark.mapper.util.RepositoryProperties;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;

/**
 * This class is a singleton managing a cache in memory for XML schemas and
 * mappings.
 */

public class MappingMetadata
{
	private static final String 	RCSRevision 	= "$Revision: 1.5 $";
	private static final String 	RCSName     	= "$Name:  $";
    
    private static HashMap mappings = new HashMap();
    private static SAXParserFactory parserFactory;
    static
    {
        parserFactory = SAXParserFactory.newInstance();
        parserFactory.setNamespaceAware(true);
        
        /* load configuration information for batch for instance */
        RepositoryProperties.load();
    }
    
    private static SchemaManager schemaManager;
    static
    {
        resetSchemas();
    }
    
    /**
     * Private constructor
     */
    private MappingMetadata()
    {}
    
    /**
     * Load a mapping file in metadata cache.
     * @param conn a JDBC connection
     * @param source the SAX InputSource encapsulating the mapping file
     * @param reload true means that the mapping file should be reloaded, even if it is present in the cache
     * @throws IOException if an IO error occured while accessing the mapping file
     * @throws SAXException if a SAX error occured while parsing the mapping file
     * @throws SQLException if an SQL error occured while accessing the database
     */
    static synchronized RepositoryMapping loadMapping(AbstractConnection conn, InputSource source, boolean reload, ClassLoader classLoader) throws IOException, SAXException, SQLException
    {
        RepositoryMapping metadata = null;
        String systemId = source.getSystemId();
        DatabaseMetaData db = conn.getConnection().getMetaData();
        String url = db.getURL();
        String user = db.getUserName();
        String key = (url != null && user != null && systemId != null) ? url+":"+user+":"+systemId : null;
        if (key != null && !reload) metadata = getMapping(key);
        if (metadata == null)
        {
            try
            {
                XMLReader reader = new ValidatingSchemaFilter(parserFactory.newSAXParser().getXMLReader(),
                new SchemaValidationContext(schemaManager));
                reader.setErrorHandler(new ValidationErrorHandler());
                Loader loader = new Loader(reader, conn, schemaManager, classLoader, false);
                MappingFactory mappingFactory = new MappingFactory(schemaManager, schemaManager);
                mappingFactory.createTree();
                metadata = loader.load(source, mappingFactory, true);
                if (key != null) storeMapping(key, metadata);
            } catch (ParserConfigurationException ex)
            {
                throw new SAXException("Could not load SAX parser", ex);
            }
        }
        return metadata;
    }
    private static RepositoryMapping getMapping(String key)
    {
        return (RepositoryMapping)mappings.get(key);
    }
    
    private static void storeMapping(String key, RepositoryMapping metadata)
    {
        mappings.put(key, metadata);
    }
    
    static SchemaManager getSchemaManager()
    {
        return schemaManager;
    }
    private static synchronized void resetSchemas()
    {
        schemaManager = new SchemaManager();
        String resource = MappingMetadata.class.getResource("/org/xquark/bridge/resources/mapping.xsd").toString();
        
        try
        {
            schemaManager.loadSchema(new InputSource(resource));
        }
        catch (SAXException e)
        {
            throw new RuntimeException("Error while loading mapping language schema.");
        }
    }
    /**
     * Loads an XML schema in memory before the mapping file loading triggers it
     * with its schemaLocation attribute (or if there is no such attribute).
     * @see org.xquark.xml.schema.SchemaManager
     */
    public synchronized static void loadSchema(InputSource source) throws SAXException
    {
        schemaManager.loadSchema(source);
    }
    
    /**
     * Empties the cache. All Mapping objects that were not closed will continue
     * working with the old metadata.
     */
    public static synchronized void reset()
    {
        mappings.clear();
        resetSchemas();
    }
    
    private static class ValidationErrorHandler implements ErrorHandler
    {
        public static final String 	RCSRevision 	= "$Revision: 1.5 $";
        public static final String 	RCSName     	= "$Name:  $";
        
        ////////////////////////////////////////////////////////////////////
        //    Implementation of ErrorHandler interface.
        ////////////////////////////////////////////////////////////////////
        public void warning(SAXParseException ex) throws SAXException
        {
            throw new SAXException("Warning at line " + ex.getLineNumber()
            + " in " + ex.getSystemId() + ": " + ex.getMessage());
        }
        
        public void error(SAXParseException ex) throws SAXException
        {
            throw new SAXException("Error at line " + ex.getLineNumber()
            + " in " + ex.getSystemId() + ": " + ex.getMessage());
        }
        
        public void fatalError(SAXParseException ex) throws SAXException
        {
            throw new SAXException("Fatal error at line " + ex.getLineNumber()
            + " in " + ex.getSystemId() + ": " + ex.getMessage());
        }
    }
}


